@@ -65,7 +65,9 @@ static struct patlist *pat_include = NULL;
 static struct patlist *pat_exclude = NULL;
 static struct range *hunks = NULL;
 static struct range *lines = NULL;
+static struct range *files = NULL;
 static enum line_numbering number_lines = None;
+static int number_files = 0;
 
 static int unzip = 0;
 static enum {
@@ -86,6 +88,7 @@ static int removing_timestamp = 0;
 static int egrepping = 0;
 static int print_patchnames = 0;
 static int empty_files_as_absent = 0;
+static unsigned long filecount=0;
 
 static int
 regexecs (regex_t *regex, size_t num_regex, const char *string,
@@ -164,13 +167,43 @@ static int output_header_line (const char *line)
 	return 0;
 }
 
+static int
+file_matches (void)
+{
+	int f = 0;
+	struct range *r;
+	
+	// See if the file range list includes this file.  -1UL is a
+	// wildcard.
+	for (r = files; r; r = r->next)
+		if ((r->start == -1UL ||
+		     r->start <= filecount) &&
+		    (r->end == -1UL ||
+		     filecount <= r->end)) {
+			f = 1;
+			break;
+		}
+
+	if (files && !f)
+		return 0;
+
+	return 1;
+}
+
 static void display_filename (unsigned long linenum, char status,
 			      const char *filename, const char *patchname)
 {
+	if (mode == mode_list && !file_matches ())
+		/* This is lsdiff --files=... and this file is not to be
+		 * listed. */
+		return;
+
 	if (print_patchnames)
 		printf ("%s:", patchname);
 	if (numbering)
 		printf ("%lu\t", linenum);
+	if (number_files)
+		printf ("File #%-3lu\t", filecount);
 	if (show_status)
 		printf ("%c ", status);
 	if (prefix_to_add)
@@ -185,6 +218,10 @@ hunk_matches (unsigned long orig_offset, unsigned long orig_count,
 	int h = 0, l = 0;
 	struct range *r;
 
+	/* The hunk can't match if the containing file doesn't. */
+	if (!file_matches ())
+		return 0;
+
 	// For the purposes of matching, zero lines at offset n counts
 	// as line n.
 	if (!orig_count)
@@ -297,12 +334,25 @@ do_unified (FILE *f, char *header[2], int match, char **line,
 							   hunknum);
 			else hunk_match = 0;
 
+			trailing = strchr (*line, '+');
+			trailing += strcspn (trailing, " ");
+			if (*trailing == ' ')
+				trailing++;
+			trailing += strspn (trailing, "@");
+
 			if (hunk_match && numbering && verbose &&
 			    mode != mode_grep) {
 				if (print_patchnames)
 					printf ("%s-", patchname);
-				printf ("\t%lu\tHunk #%lu\n",
+				printf ("\t%lu\tHunk #%lu",
 					hunk_linenum, hunknum);
+				if (verbose > 1) {
+					char *p = trailing;
+					if (*p != '\n')
+						p++;
+					printf ("\t%s", p);
+				} else
+					putchar ('\n');
 			}
 
 			if (hunk_match &&
@@ -349,11 +399,6 @@ do_unified (FILE *f, char *header[2], int match, char **line,
 							 " Hunk #%lu, %s",
 							 hunknum, bestname);
 
-					trailing = strchr (*line, '+');
-					trailing += strcspn (trailing, " ");
-					if (*trailing == ' ')
-						trailing++;
-					trailing += strspn (trailing, "@");
 					fputs (trailing, output_to);
 					break;
 				case Before:
@@ -383,10 +428,12 @@ do_unified (FILE *f, char *header[2], int match, char **line,
 			continue;
 		}
 
-		if (orig_count && **line != '+')
-			orig_count--;
-		if (new_count && **line != '-')
-			new_count--;
+		if (**line != '\\') {
+			if (orig_count && **line != '+')
+				orig_count--;
+			if (new_count && **line != '-')
+				new_count--;
+		}
 
 		if (hunk_match && mode == mode_grep &&
 		    !regexecs (regex, num_regex, *line + 1, 0, NULL, 0)) {
@@ -892,6 +939,7 @@ static int filterdiff (FILE *f, const char *patchname)
 			continue;
 		}
 
+		filecount++;
 		header[1] = xstrdup (line);
 		names[1] = filename_from_header (line + 4);
 
@@ -968,19 +1016,22 @@ const char * syntax_str =
 "  --hunks=H, -# H\n"
 "            include only hunks in range H\n"
 "  --lines=L include only hunks with (original) lines in range L\n"
+"  --files=F include only files in range F\n"
 "  --annotate (filterdiff, grepdiff)\n"
 "            annotate each hunk with the filename and hunk number (filterdiff, grepdiff)\n"
 "  --as-numbered-lines=before|after (filterdiff, grepdiff)\n"
 "            display lines as they would look before, or after, the (filterdiff, grepdiff)\n"
 "            patch is applied (filterdiff, grepdiff)\n"
 "  --format=context|unified (filterdiff, grepdiff)\n"
-"            set output format\n"
+"            set output format (filterdiff, grepdiff)\n"
 "  --output-matching=hunk|file (grepdiff)\n"
 "            show matching hunks or file-level diffs (grepdiff)\n"
 "  --remove-timestamps (filterdiff, grepdiff)\n"
 "            don't show timestamps from output (filterdiff, grepdiff)\n"
 "  -z        decompress .gz and .bz2 files\n"
-"  -n        show line numbers (lsdiff)\n"
+"  -n        show line numbers (lsdiff, grepdiff)\n"
+"  --number-files (lsdiff, grepdiff)\n"
+"            show file numbers, for use with filterdiff's --files option (lsdiff, grepdiff)\n"
 "  -H, --with-filename (lsdiff, grepdiff)\n"
 "            show patch file names (lsdiff, grepdiff)\n"
 "  -h, --no-filename (lsdiff, grepdiff)\n"
@@ -990,7 +1041,7 @@ const char * syntax_str =
 "  --addprefix=PREFIX\n"
 "            prefix pathnames with PREFIX\n"
 "  -s        show file additions and removals (lsdiff)\n"
-"  -v        verbose output\n"
+"  -v        verbose output -- use more than once for extra verbosity\n"
 "  -E        use extended regexps, like egrep (grepdiff)\n"
 "  -E, --empty-files-as-absent (lsdiff)\n"
 "            treat empty files as absent (lsdiff)\n"
@@ -1194,6 +1245,7 @@ int main (int argc, char *argv[])
 			{"addprefix", 1, 0, 1000 + 'A'},
 			{"hunks", 1, 0, '#'},
 			{"lines", 1, 0, 1000 + ':'},
+			{"files", 1, 0, 1000 + 'w'},
 			{"as-numbered-lines", 1, 0, 1000 + 'L'},
 			{"annotate", 0, 0, 1000 + 'a'},
 			{"format", 1, 0, 1000 + 'F'},
@@ -1202,6 +1254,7 @@ int main (int argc, char *argv[])
 			{"with-filename", 0, 0, 'H'},
 			{"no-filename", 0, 0, 'h'},
 			{"empty-files-as-absent", 0, 0, 'E'},
+			{"number-files", 0, 0, 1000 + 'n'},
 			{0, 0, 0, 0}
 		};
 		char *end;
@@ -1265,11 +1318,16 @@ int main (int argc, char *argv[])
 		case 'n':
 			numbering = 1;
 			break;
+		case 1000 + 'n':
+			number_files = 1;
+			break;
 		case 's':
 			show_status = 1;
 			break;
 		case 'v':
 			verbose++;
+			if (numbering && verbose > 1)
+				number_files = 1;
 			break;
 		case '#':
 			if (hunks)
@@ -1292,6 +1350,11 @@ int main (int argc, char *argv[])
 				syntax (1);
 			parse_range (&lines, optarg);
 			break;
+		case 1000 + 'w':
+			if (files)
+				syntax (1);
+			parse_range (&files, optarg);
+			break;
 		case 1000 + 'L':
 			if (!strcmp (optarg, "before"))
 				number_lines = Before;
@@ -1342,7 +1405,7 @@ int main (int argc, char *argv[])
 
 	if (numbering &&
 	    !(mode == mode_list ||
-	      output_matching == output_none))
+	      (mode == mode_grep && output_matching == output_none)))
 		error (EXIT_FAILURE, 0, "-n only applies to list mode");
 
 	if (mode != mode_filter &&
